#ifndef _DRAW_H
#define _DRAW_H


#include "colors.h"



//-------------------------------------------------------
// Display list per i vari
// ordinamenti spaziali degli slices
typedef enum
{
 FRONT_BACK = 0,
 BACK_FRONT,
 UP_DOWN,
 DOWN_UP,
 LEFT_RIGHT,
 RIGHT_LEFT
} t_enum_orientation;


GLuint g_orientation_lists[6][256];
int g_dim_ori_lists[6] = {DIM_Z,DIM_Z, DIM_Y,DIM_Y, DIM_X,DIM_X};

GLfloat g_ori_normals[6][3] =
{
	{ 1, 0, 0 }, // FRONT_BACK
	{-1, 0, 0 }, // BACK_FRONT
	{ 0, 1, 0 }, // UP_DOWN
	{ 0,-1, 0 }, // DOWN_UP
	{ 0, 0,-1 }, // LEFT_RIGHT
	{ 0, 0, 1 }, // RIGHT_LEFT
};


//------------------------------------------------------------------------------


t_enum_orientation get_orientation()
{
return FRONT_BACK;


static float axes[4][3] =
{
	{ 1, 0, 0 }, //      \    XY
	{ 0, 1, 0 }, //       /
	{ 0, 0, 1 }, //      V
};


float x, y, z;
float ox, oy, oz;
float v[3];
static float sqrt2_2 = sqrt(2.0) / 2.0;
float ta, tb;

int i,j;


// trasforma l'origine:
glTransform(0,0,0, &ox, &oy, &oz);
// trasforma il punto (0,1,0)
glTransform(0,1,0, &x, &y, &z);

v[0] = x - ox;
v[1] = y - oy;
v[2] = z - oz;

int code = 0x00;
for(j=0; j<3; j++)
{
float t = 0.0;
 for(i=0; i<3; i++)
 {
	 t +=v [i] * axes[j][i];
 }
 if(t > 0.0) code |= (1<<j);
}

return (t_enum_orientation)code;



}





void draw_slice(t_enum_orientation ori, int num_slice, float *data, int scale)
{
int i, j;
float a,b;
float da,db;
int na,nb;
int dpa,dpb, off_1,off_2,off_3;
float *ptr;

if(num_slice < 0) num_slice = 0; 
 else
if(num_slice > g_dim_ori_lists[ori] - 1)  num_slice = g_dim_ori_lists[ori] - 1;



glPushMatrix();


ptr = data;

// Calcola l'offset iniziale nei dati:
if(ori == FRONT_BACK || ori == BACK_FRONT)
{
  ptr += DIM_X*DIM_Y * num_slice;

	na = DIM_X/scale-1;
  nb = DIM_Y/scale-1;

  da =  2.0 / DIM_X;
  db =  2.0 / DIM_Y;

  dpa = scale;
  dpb = dpa + DIM_X*(scale-1);

  glTranslatef(0,0,(num_slice*2.0)/(float)DIM_Z-1.0);
	//glRotatef();

  off_1 = scale;
  off_2 = DIM_X*scale+scale;
  off_3 = DIM_X*scale;

  b=-1.0;
}

 else

if(ori == UP_DOWN || ori == DOWN_UP)  
{
  ptr += DIM_X * num_slice;

  na = DIM_X/scale-1;
  nb = DIM_Z/scale-1;

	da =  2.0 / DIM_X;
  db =  2.0 / DIM_Z;

  dpa = scale;
  dpb = dpa + DIM_X*DIM_Y*(scale);


	
	glTranslatef(0,(num_slice*2.0)/(float)DIM_Y-1.0,0);
  glRotatef(90.0,1,0,0);

  off_1 = scale;
  off_2 = DIM_X*DIM_Y*scale + scale;
  off_3 = DIM_X*DIM_Y*scale;

	b=-1.0;
}

 else

if(ori == LEFT_RIGHT || ori == RIGHT_LEFT)
{
  ptr += num_slice;

  na = DIM_Y/scale-1;
  nb = DIM_Z/scale-1;

  da =   2.0 / DIM_Y;
  db =   2.0 / DIM_Z;

  dpa = DIM_X*scale;
  dpb = dpa + DIM_X*DIM_Y*(scale-1);


//	glTranslatef((num_slice*2.0)/(float)DIM_X-1.0,0,0);
  glRotatef(90, 0,0,1);
	glRotatef(90, 1,0,0);
  glTranslatef(0,0,(num_slice*2.0)/(float)DIM_X-1.0);

  off_1 = scale*DIM_X;
  off_2 = scale*DIM_X + DIM_X*DIM_Y*scale;
  off_3 = DIM_X*DIM_Y*scale;

	b=-1.0;
}


da *= (float)scale;
db *= (float)scale;


glBegin(GL_QUADS);


glNormal3f( 0,0,1 );




for(j=0; j < nb; j++, ptr+=dpb)
{

 a=-1.0;

 for(i=0; i < na; i++, ptr+=dpa)
 {
  
   //glColor4fv( get_palette(ptr[0]) );
	 glTexCoord1f(ptr[0]);
   glVertex2f(a,b);

   //glColor4fv( get_palette(ptr[scale]) );
	 glTexCoord1f(ptr[off_1]);
   glVertex2f(a+da,b);

   //glColor4fv( get_palette(ptr[0+offs]) );
	 glTexCoord1f(ptr[off_2]);
   glVertex2f(a+da,b+db);

   //glColor4fv( get_palette(ptr[scale+offs]) );
	 glTexCoord1f(ptr[off_3]);
   glVertex2f(a,b+db);

   a+=da;
 }
 b+=db;
 

}



glEnd();
 

glPopMatrix();

}








void draw_volume(int scale, float rot_x, float rot_y)
{

int ori = ((int)get_orientation())%6;

if(g_orientation_lists[ori][0])
{
	//glCallLists( GLsizei n, GLenum type, const GLvoid *lists )
	glCallLists( g_dim_ori_lists[ori], GL_UNSIGNED_INT, (GLvoid *)g_orientation_lists[ori] );
}

else

for(ori=0; ori<6; ori+=2)
{
printf("Calcolando le Display List... ");

GLuint base = glGenLists( g_dim_ori_lists[ori] );

 for(int i=0; i < g_dim_ori_lists[ori]; i++)
 {
  g_orientation_lists[ori][i] = base++;

  glNewList( g_orientation_lists[ori][i], GL_COMPILE_AND_EXECUTE );

      draw_slice((t_enum_orientation)ori, i, g_grid_vals, scale );

  glEndList();
 }

for(i=0; i < g_dim_ori_lists[ori]; i++)
	g_orientation_lists[ori+1][g_dim_ori_lists[ori]-i-1] = g_orientation_lists[ori][i];

printf("fatto :)\n");

}


}










GLuint g_display_list2 = 0;
float  g_my_iso_m = -1.0;
float  g_my_iso = .3;

void draw_iso(int scale, float iso)
{


if(g_display_list2 && iso == g_my_iso_m)
{
	glCallList(g_display_list2);
}


else
{


printf("\n(iso = %.4f) Calcolando la Display List... ", iso);

if(g_my_iso_m == -1.0) g_display_list2 = glGenLists( 1 );
g_my_iso_m = iso;
glNewList( g_display_list2, GL_COMPILE_AND_EXECUTE );

int poly =
 iso_draw_surface(0.0,0.0,0.0,
	                DIM_X,DIM_Y,DIM_Z,
									2.0,2.0,2.0, iso);


glEndList();
printf("fatto :) [%i triangoli]\n", poly);
}

}










#endif
