////////////////////////////////////////////////////
//   *** MARCHING CUBES ALGORITHM LIBRARY ***     //
//                                                //
// Written by Paul Bourke						              //
// May 1997										                    //
//												                        //
// Based on tables by Cory Gene Bloyd			        //
//												                        //
// Enhanced by Bosi Mathieu						            //
//                                                //
////////////////////////////////////////////////////
////////////////////////////////////////////////////


#ifndef _MARCHING_CUBES_LIBRARY_
#define _MARCHING_CUBES_LIBRARY_ 

#include "vectors3d.h"
#include "marching_list.h"  // contiene g_edge_table[] e g_iso_tri_table[]



//#define ABS( a )  ((a)<0?(-a):(a))
#define POW2( a ) ((a) * (a))

#define GET_VAL_ISO( v_i, v_f, v_iso)   (((v_iso) - (v_i)) / ((v_f) - (v_i)))

#define GRAD_DELTA (1.0/64.0)

// #define:





// numerical type used for the 3D field-function value storage
#define GRID_TYPE float 


#define GRID_X 2.0
#define GRID_Y 2.0 
#define GRID_Z 2.0



// #define per il calcolo dei Julia3D:
#define QUAT_JULIA_BAIL_OUT 2.0
#define QUAT_JULIA_MAX_ITER 32
#define QUAT_JULIA_REC_MAX_ITER (1.0/(float)(QUAT_JULIA_MAX_ITER))
//#define QUAT_JULIA_MANDELBROT


//#define FIELD_FUN eval_quat_julia
#define FIELD_FUN eval_field




//
//   Vertex numbering 
//   in Paul Bourke's algorithm:
//  
//       _____4______
//    7/|           5|
//  off____6______ / |
//   |            |  |
//  11  8        10  9
//   |  |____0___ | _|
//   | 3/         |  / 1
//   |/___________|/
//          2      
//




// tipo che definisce una meta-ball
typedef struct
{
 float x, y, z;        // posizione del centro all'inside griglia
 float r, r2, rec_r2;  // radius, quadrato del radius, reciproco del quadrato del radius
} t_meta_ball;

// numero di MetaBalls da considerare
#define NUM_BALLS 4

// struttura dati per le MetaBalls
t_meta_ball g_meta_ball[NUM_BALLS];


GRID_TYPE  g_my_iso_value = 0.4;





//
// PROTOTYPES::
//


//__inline float VertexInterp(t_point3d *dest, float isolevel, GridVert *p1, GridVert *p2);
//__inline void Polygonise(Grid3D *g, int num_p, float isolevel);


void metaballs_init(void);
void metaballs_update(float aa);




float eval_field(float x, float y, float z);
float eval_quat_julia(float x, float y, float z);

void  eval_gradient(t_vector3d *pos, t_vector3d *res);

void draw_isocurve(float iso, t_basis3d *b, float dim_u, float dim_v, int nu, int nv);

int  iso_draw_surface( float ox, float oy, float oz,  int nu, int nv, int nw,
					             float dim_u, float dim_v, float dim_w, float iso);





/***************************************/





///////////////////////////////////////////////////
// Funzioni analitiche per il calcolo            //
// del campo e del gradiente in un punto (x,y,z) //
///////////////////////////////////////////////////
float eval_field(float x, float y, float z)
{
static float dens, dist2, t;
int k;
static t_meta_ball *mb; 

   dens = 0.0;

   for(k=0; k<NUM_BALLS; k++)
   {
    mb = &g_meta_ball[k];

     t = mb->x - x; if(t<0) t=-t; if(t > mb->r) continue;
	 t = mb->y - y; if(t<0) t=-t; if(t > mb->r) continue;
	 t = mb->z - z; if(t<0) t=-t; if(t > mb->r) continue;

     {
	  
      t = mb->x - x; dist2=t*t;
      t = mb->y - y; dist2+=t*t;
      t = mb->z - z; dist2+=t*t;

      if( dist2 > mb->r2 ) continue;

      {
	  //dens = pow(1.0-(dist2 * mb->rec_r2), 4);
	   t = 1.0 - (dist2 * mb->rec_r2);
	   t *= t;
        // if(mb->neg == true) dens -= tmp*tmp;  else
         dens += t*t;
		 
      }
	}
   } // end for(;;)


/*
// Colonna:

   dens+= fabs(y*.5) / (POW2(xx*2.0) + POW2(zz*2.0) + 1.0) * 1.5;
   dens+= fabs(x*.5) / (POW2(yy*2.0) + POW2(zz*2.0) + 1.0) * 1.5;
   dens+= fabs(z*.5) / (POW2(xx*2.0) + POW2(yy*2.0) + 1.0) * 1.5;
*/
/*   
   t=x*x+y*y;
   dens+=.3/(t*t+1.0);
   t=x*x+z*z;
   dens+=.3/(t*t+1.0);
   t=z*z+y*y;
   dens+=.3/(t*t+1.0);
*/
// Pareti sup. e inf.
/*   
#define LARG 0.8
   if(!(x>-LARG && x<LARG &&
        y>-LARG && y<LARG &&
        z>-LARG && z<LARG )) dens=.5;
		#undef LARG
*/




return dens;

}
///////////////////////////////////////////////////////////////////////////////
// calcola il vettore gradiente nel punto 'pos'  tramite differenze centrali //
///////////////////////////////////////////////////////////////////////////////
void  eval_gradient(t_vector3d *pos,  t_vector3d *res)
{
    res->x  =  FIELD_FUN(pos->x - GRAD_DELTA, pos->y, pos->z) - 
		           FIELD_FUN(pos->x + GRAD_DELTA, pos->y, pos->z);

    res->y  =  FIELD_FUN(pos->x, pos->y - GRAD_DELTA, pos->z) - 
		           FIELD_FUN(pos->x, pos->y + GRAD_DELTA, pos->z);

    res->z  =  FIELD_FUN(pos->x, pos->y, pos->z - GRAD_DELTA) - 
		           FIELD_FUN(pos->x, pos->y, pos->z + GRAD_DELTA);	

    vector3d_norm( res );
}
/////////////////////




/////////////////////////////////////////////////////////
// (funzioni per il rendering di linee equipotenziali) //
/////////////////////////////////////////////////////////
/********************************************************/
//   Numerazione vertici-spigoli cella 2D:
//   2---P2---3
//   |        |
//   P1       P3
//   |        |
//   1---P4---4


void draw_isocurve(float iso, t_basis3d *b, float dim_u, float dim_v, int nu, int nv)
{
float xu, yu, zu;
float dxu, dyu, dzu;
float xv, yv, zv;
float dxv, dyv, dzv;
float du, dv;
int i, j;
t_point3d p1, p2, p3, p4;
float va, vb, vc, vd;
int index, f;
float t;

//variabili di look-up
static t_point3d g_mem_p3[256];
static float mem_vcd[256];
float  dxu_dxv, dyu_dyv, dzu_dzv;

//if(iso2D_look_up_initialized==false)  init_iso_val();
{
t_point3d p0={0,0,0};
p1=p2=p3=p4=p0;
}

du = dim_u / (float)nu;
dv = dim_v / (float)nv;

// ???
// (normalizzare o no i vettori della base?
//  forse e' limitante per basi non normalizzate)

// g_arr_delta[ per spostarsi nella direzione u
dxu = b->u.x * du;
dyu = b->u.y * du;
dzu = b->u.z * du;
// g_arr_delta[ per spostarsi nella direzione v
dxv = b->v.x * dv;
dyv = b->v.y * dv;
dzv = b->v.z * dv;


// posiziona il punto di partenza u(xu,yu,zu) sull'origine della base
xu = b->o.x;
yu = b->o.y;
zu = b->o.z; 

// per centrare l'origine
xu -= (dim_u*.5) * b->u.x + (dim_v*.5) * b->v.x;
yu -= (dim_u*.5) * b->u.y + (dim_v*.5) * b->v.y;
zu -= (dim_u*.5) * b->u.z + (dim_v*.5) * b->v.z;



dxu_dxv = dxu + dxv;
dyu_dyv = dyu + dyv;
dzu_dzv = dzu + dzv;


glBegin(GL_LINES);

for(i=0; i<nu; i++) // fa variare il punto di partenza sull'asse u
{
 xv = xu;
 yv = yu;
 zv = zu; 

 for(j=0; j<nv; j++) // dal punto di partenza sull'asse u va parallelamente a v
 {
  /************/

//   Numerazione vertici-spigoli cella 2D:
//
//  ^|        |
//   b---P2---c- 
//   |        |
//   P1       P3
//   |        |
//   a---P4---d-


if(i)  // coerenza orizzontale
{
  if(j)
  {
   va = vb;
   vb = mem_vcd[j+1];
  }
  else
  {
   va = mem_vcd[j];
   vb = mem_vcd[j+1];
  }
}
else
{
	 // calcola 'va' e 'vb'
	 if(j)
	 {
	  va = vb;
	  vb = FIELD_FUN(xv + dxv, yv + dyv, zv + dzv);
	 }
	 else 
	 {	 
	  va = FIELD_FUN(xv, yv, zv);
	  vb = FIELD_FUN(xv + dxv, yv + dyv, zv + dzv);
	 }
}



// calcola 'vc' e 'vd'
if(j)
{
 vd = vc;
 //vc = eval_field(xv + dxu + dxv, yv + dyu + dyv, zv + dzu + dzv);
 vc = FIELD_FUN(xv + dxu_dxv, yv + dyu_dyv, zv + dzu_dzv);
 mem_vcd[j+1] = vc;
}
else
{
 vd = FIELD_FUN(xv + dxu, yv + dyu, zv + dzu);
 vc = FIELD_FUN(xv + dxu_dxv, yv + dyu_dyv, zv + dzu_dzv);
 mem_vcd[j]   = vd;
 mem_vcd[j+1] = vc;
}



  /************/



  // ricostruisce l'indice di accesso in base ai valori rilevati ai 4 vertici 
  // if(j)  
  index = 0x0000;
   if(va > iso)  index|=0x0001;
   if(vb > iso)  index|=0x0002;
   if(vc > iso)  index|=0x0004;
   if(vd > iso)  index|=0x0008;

  
  //iso2D_look_up[index] mantiene codificati al suo inside gli spigoli "attivi"

  f = iso2D_look_up[index];   
  if( f )// continue;// se f!=0 -> la cella e' attraversata dalla superficie equipotenziale  
  {

//  bit 0-3: indice edge-points (Pn) da calcolare   [se alti]
//   Numerazione vertici-spigoli cella 2D:
//
//   |        |
//   2---P2---3- 
//   |        |
//   P1       P3
//   |        |
//   1---P4---4-

// val_iso = (v_iso - v_i) / (v_f - v_i)


// (coerenza verticale)
if(i)
{
 
  if(f & 1) p1 = g_mem_p3[j];

  if(f & 4) // calcola P3 (verticale)
  {
   t = GET_VAL_ISO(vd, vc, iso);
   p3.x = xv + dxu  +  t*dxv;
   p3.y = yv + dyu  +  t*dyv;
   p3.z = zv + dzu  +  t*dzv;

   g_mem_p3[j] = p3;
  }
  
}
else
{
  if(f & 1) // calcola P1 (verticale)
  {
   t = GET_VAL_ISO(va, vb, iso);
   p1.x = xv  +  t*dxv;
   p1.y = yv  +  t*dyv;
   p1.z = zv  +  t*dzv;
  }
  if(f & 4) // calcola P3 (verticale)
  {
   t = GET_VAL_ISO(vd, vc, iso);
   p3.x = xv + dxu  +  t*dxv;
   p3.y = yv + dyu  +  t*dyv;
   p3.z = zv + dzu  +  t*dzv;
   g_mem_p3[j] = p3;
  }
}


// (coerenza orizzontale)
if(j)
{
  if(f & 8)  p4 = p2;

  if(f & 2) // calcola P2 (orizzontale)
  {
   t = GET_VAL_ISO(vb, vc, iso);
   p2.x = xv + dxv  +  t*dxu;
   p2.y = yv + dyv  +  t*dyu;
   p2.z = zv + dzv  +  t*dzu;
  }
}
else
{
  if(f & 2) // calcola P2 (orizzontale)
  {
   t = GET_VAL_ISO(vb, vc, iso);
   p2.x = xv + dxv  +  t*dxu;
   p2.y = yv + dyv  +  t*dyu;
   p2.z = zv + dzv  +  t*dzu;
  }
  if(f & 8) // calcola P4 (orizzontale)
  {
   t = GET_VAL_ISO(va, vd, iso);
   p4.x = xv + t*dxu;
   p4.y = yv + t*dyu;
   p4.z = zv + t*dzu;
  }
}

/******/


   //  bit 4:   si tratta di spigoli isolati(non contigui) [se alto]
   if( f & 0x0010 ) // (alto -> spigoli isolati, procedi senza problemi)
	{	  
	 if( f & 0x0100 ) { glVertex3fv( (float *)&p4.x ); glVertex3fv( (float *)&p1.x ); }
     if( f & 0x0200 ) { glVertex3fv( (float *)&p1.x ); glVertex3fv( (float *)&p2.x ); }
     if( f & 0x0400 ) { glVertex3fv( (float *)&p2.x ); glVertex3fv( (float *)&p3.x ); }
     if( f & 0x0800 ) { glVertex3fv( (float *)&p3.x ); glVertex3fv( (float *)&p4.x ); } 	 
	}
   
   else  // altrimenti per forza spigoli contigui (orizzontali o verticali)
   // bit 5: 0 -> 1100  orizzontale P1->P3
   //        1 -> 1001  verticale   P2->P4
   if(f & 32) // verticale
	{ glVertex3fv( (float *)&p2.x ); glVertex3fv( (float *)&p4.x ); }
   else       // orizzontale
	{ glVertex3fv( (float *)&p1.x ); glVertex3fv( (float *)&p3.x ); }
     

  } // if(f)

	
  // si sposta parallelamente all'asse V
  xv+=dxv;
  yv+=dyv;
  zv+=dzv;  
 }// for(j=0; j<nv; j++)

 // si sposta lungo l'asse U
 xu+=dxu;
 yu+=dyu;
 zu+=dzu;

}// for(i=0; i<nu; i++)

glEnd();

}







/*
   Linearly interpolatete the position where an isosurface cuts
   an edge between two vertices, each with their own scalar value
*/
/*
__inline float VertexInterp(t_point3d *dest, float isolevel, GridVert *p1, GridVert *p2)
{
   float mu, t;
   static t_point3d *p;

   mu = (isolevel - p1->val) / (float)(p2->val - p1->val);

   p = &p1->p;   

   dest->x = p->x  +  mu * (p2->p.x - p->x);
   dest->y = p->y  +  mu * (p2->p.y - p->y);
   dest->z = p->z  +  mu * (p2->p.z - p->z);

return mu;

}
*/



int g_num_iso_cells=0;




//
// Routines per il calcolo delle meta-balls:
//



//(stdcall *)(GRID_TYPE grid_val_func(int, int, int));




void metaballs_init(void)
{
int i;

 g_meta_ball[0].r = GRID_X/3.0;
 g_meta_ball[1].r = GRID_X/2.5;
 g_meta_ball[2].r = GRID_X/2.75;
 g_meta_ball[3].r = GRID_X/2.85;


// g_meta_ball[2].neg=true;

  for(i=0; i<NUM_BALLS; i++) 
  {
    g_meta_ball[i].r2 = pow(g_meta_ball[i].r,2);
	g_meta_ball[i].rec_r2 = 1.0 / g_meta_ball[i].r2;
  }
}






void metaballs_update(float aa)
{
 g_meta_ball[0].x = sin(aa*3.5)*GRID_X/6.0;
 g_meta_ball[0].y = cos(aa*2.0)*GRID_Y/6.0;
 g_meta_ball[0].z = sin(aa)*GRID_Z/8.0;

 g_meta_ball[1].x = sin(aa)*GRID_X/4.0;
 g_meta_ball[1].y = sin(aa*2.6*2.0)*GRID_Y/4.0;
 g_meta_ball[1].z = sin(aa*0.63)*GRID_Z/8.0;

 g_meta_ball[2].x = sin(aa)*GRID_X/8.0;
 g_meta_ball[2].y = -sin(aa*1.3*2.0)*GRID_Y/4.3;
 g_meta_ball[2].z = GRID_Z/4.0;

 g_meta_ball[3].x = sin(aa*1.3*2.0)*GRID_Y/4.3;
 g_meta_ball[3].y = -cos(aa) * GRID_X/8.0;
 g_meta_ball[3].z = sin(aa*0.67)*GRID_Z/4.0;
}
















static float QUAT_JULIA_Kewl[5][5] =
{
 { -0.08, 0.0, -0.8, -0.01, 0.0 },

 { -0.75, 0.023, 0.12, 0.0, 0.0 },

 { -1.25, 0.0, 0.0, 0.0, 0.0 },

 { -0.125, -0.256, 0.847, 0.0895, 0.0 },

 { -0.213, -0.0410, -0.563, -0.560, 0.0 }
};







float
 QUAT_JULIA_ox = -1.25,
 QUAT_JULIA_oy = -0.256,
 QUAT_JULIA_oz = 0.847,
 QUAT_JULIA_ow = 0.0895,
 QUAT_JULIA_ot = 0.0;

void set_Julia3D(float ox, float oy, float oz, float ow, float ot)
{
 QUAT_JULIA_ox = ox;
 QUAT_JULIA_oy = oy;
 QUAT_JULIA_oz = oz;
 QUAT_JULIA_ow = ow;
 QUAT_JULIA_ot = ot;
}


void set_Julia3D_preset(int n)
{
n%=5;
 QUAT_JULIA_ox = QUAT_JULIA_Kewl[0][n];
 QUAT_JULIA_oy = QUAT_JULIA_Kewl[1][n];
 QUAT_JULIA_oz = QUAT_JULIA_Kewl[2][n];
 QUAT_JULIA_ow = QUAT_JULIA_Kewl[3][n];
 QUAT_JULIA_ot = QUAT_JULIA_Kewl[4][n];
}


void morph_Julia3D(int n, float t)
{
n%=5;

set_Julia3D(
QUAT_JULIA_Kewl[n][0] + sin(t)*.4,
QUAT_JULIA_Kewl[n][1],
QUAT_JULIA_Kewl[n][2],
QUAT_JULIA_Kewl[n][3],
QUAT_JULIA_Kewl[n][4] );
}

float eval_quat_julia(float x, float y, float z)
{
float w;
float a, b, c, d;
float ta, tb, tc, td;

int k;

float 
tmp_00, tmp_01, tmp_02, 
tmp_03, tmp_04, tmp_05, 
tmp_06, tmp_07, tmp_08, tmp_09;


#ifndef QUAT_JULIA_MANDELBROT
    a = x;
    b = y;
    c = z;
    d = QUAT_JULIA_ot; 
#endif
#ifdef QUAT_JULIA_MANDELBROT
 a = QUAT_JULIA_ox;
 b = QUAT_JULIA_oy;
 c = QUAT_JULIA_oz;
 d = QUAT_JULIA_ot;
#endif

     k = 0;

	  while(1)
	  {

	   if( k >= QUAT_JULIA_MAX_ITER ) break;

       if( a*a + b*b + c*c + d*d >= POW2( QUAT_JULIA_BAIL_OUT ) ) break;	   

/*
Quaternions multiplication (optimized method):

   q = (a, b, c, d), p = (x, y, z, w)

   tmp_00 = (d - c) * (z - w)
   tmp_01 = (a + b) * (x + y)
   tmp_02 = (a - b) * (z + w)
   tmp_03 = (c + d) * (x - y)
   tmp_04 = (d - b) * (y - z)
   tmp_05 = (d + b) * (y + z)
   tmp_06 = (a + c) * (x - w)
   tmp_07 = (a - c) * (x + w)
   tmp_08 = tmp_05 + tmp_06 + tmp_07
   tmp_09 = 0.5 * (tmp_04 + tmp_08)

   q * p = (tmp_00 + tmp_09 - tmp_05,
            tmp_01 + tmp_09 - tmp_08,
            tmp_02 + tmp_09 - tmp_07,
            tmp_03 + tmp_09 - tmp_06)


////////////////////////////////////////

thanks to Lord Hamilton, who invented quaternions 
and to Henrik Engstrm, who wrote a cool doc about them in 1994-Oct-27
*/

/* q = (a, b, c, d):  questo sotto e' q^2 */

		{
		 tmp_00 = c - d; 
		  tmp_00 *= -tmp_00;

		 tmp_01 = a + b; 
		  tmp_01 *= tmp_01;

		 tmp_03 = tmp_02 = (a - b) * (c + d);
   
	     tmp_05 = (d + b) * (b + c);
		 tmp_06 = (a + c) * (a - d);
		 tmp_07 = (a - c) * (a + d);
		 tmp_08 = tmp_05 + tmp_06 + tmp_07;

		 tmp_04 = (d - b) * (b - c);   
		 tmp_09 = 0.5 * (tmp_04 + tmp_08);

		/* somma la costante  c(QUAT_JULIA_ox,QUAT_JULIA_oy,QUAT_JULIA_oz,QUAT_JULIA_ow) */
		#ifndef QUAT_JULIA_MANDELBROT
		  ta  =  (tmp_00  +  tmp_09  -  tmp_05)  +  QUAT_JULIA_ox;
		  tb  =  (tmp_01  +  tmp_09  -  tmp_08)  +  QUAT_JULIA_oy; 
		  tc  =  (tmp_02  +  tmp_09  -  tmp_07)  +  QUAT_JULIA_oz;
		  td  =  (tmp_03  +  tmp_09  -  tmp_06)  +  QUAT_JULIA_ow;
		#endif
		#ifdef QUAT_JULIA_MANDELBROT
		  ta  =  (tmp_00  +  tmp_09  -  tmp_05)  +  x;
		  tb  =  (tmp_01  +  tmp_09  -  tmp_08)  +  y; 
		  tc  =  (tmp_02  +  tmp_09  -  tmp_07)  +  z;
		  td  =  (tmp_03  +  tmp_09  -  tmp_06)  +  QUAT_JULIA_ot;
		#endif

		  a = ta;
		  b = tb;
		  c = tc;
		  d = td;
		}
       
	   k++;

	  } // fine del blocco ricorsivo while(1)

  
   return ( k * QUAT_JULIA_REC_MAX_ITER );
}





//********************************************//

#define GRID_DIM DIM_X


// FIELD FUNCTION XY STORAGE ARRAYS:

float g_mem_block[GRID_DIM*GRID_DIM * 4];

float *g_field_0 = g_mem_block + GRID_DIM*GRID_DIM*0;
float *g_field_1 = g_mem_block + GRID_DIM*GRID_DIM*1;
float *g_field_2 = g_mem_block + GRID_DIM*GRID_DIM*2;
float *g_field_3 = g_mem_block + GRID_DIM*GRID_DIM*3;


// VERTICES: FUNCTION GRADIENT STORAGE ARRAYS:

t_vector3d g_mem_block1[GRID_DIM*GRID_DIM * 4];

t_vector3d 
 *grv_0 = g_mem_block1 + GRID_DIM*GRID_DIM*0,
 *grv_1 = g_mem_block1 + GRID_DIM*GRID_DIM*1,
 *grv_2 = g_mem_block1 + GRID_DIM*GRID_DIM*2,
 *grv_3 = g_mem_block1 + GRID_DIM*GRID_DIM*3;



// GRADIENT AND VERTEX STORAGE ARRAYS:

//  EDGE NUMBERING:
//       _____4______
//    7/|           5|
//  off____6______ / |
//   |            |  |
//  11  8        10  9 
//   |  |___0____ | _|
//   | 3/         |  / 1 
//   |/___________|/
//          2      
//   note: edges 0, 1 and 9 will be computed for each cell
//   



t_vector3d g_mem_block2[ (GRID_DIM*GRID_DIM) * 8   +  (GRID_DIM) * 6 ];


t_vector3d 
 *g_mem_p0_p = g_mem_block2 + GRID_DIM*GRID_DIM*0,
 *g_mem_p4_p = g_mem_block2 + GRID_DIM*GRID_DIM*1,
 *g_mem_p8_p = g_mem_block2 + GRID_DIM*GRID_DIM*2,
 *g_mem_p9_p = g_mem_block2 + GRID_DIM*GRID_DIM*3;
t_vector3d
 *g_mem_gr0_p = g_mem_block2 + GRID_DIM*GRID_DIM*4,
 *g_mem_gr4_p = g_mem_block2 + GRID_DIM*GRID_DIM*5,
 *g_mem_gr8_p = g_mem_block2 + GRID_DIM*GRID_DIM*6,
 *g_mem_gr9_p = g_mem_block2 + GRID_DIM*GRID_DIM*7;


t_vector3d
 *g_mem_p1 = g_mem_block2 + ((GRID_DIM*GRID_DIM) * 8) + GRID_DIM*0,
 *g_mem_p2 = g_mem_block2 + ((GRID_DIM*GRID_DIM) * 8) + GRID_DIM*1,
 *g_mem_p3 = g_mem_block2 + ((GRID_DIM*GRID_DIM) * 8) + GRID_DIM*2;
t_vector3d
 *g_mem_gr1 = g_mem_block2 + ((GRID_DIM*GRID_DIM) * 8) + GRID_DIM*3, 
 *g_mem_gr2 = g_mem_block2 + ((GRID_DIM*GRID_DIM) * 8) + GRID_DIM*4,
 *g_mem_gr3 = g_mem_block2 + ((GRID_DIM*GRID_DIM) * 8) + GRID_DIM*5;



//////////////////////////////////////




int  iso_draw_surface( float ox, float oy, float oz, 
					             int nu, int nv, int nw,
					             float dim_u, float dim_v, float dim_w, float iso)
{
// flag di accesso alle tabelle di look-up
   int   cubeindex;
   short edge_table_val;   

// contatori:
int ix, iy, iz;
int   i, num_tri=0, ind;

// variabili vertici:
float     v[8]; // valori del campo agli 8 vertici

t_vector3d *grv[8]; // gradienti agli 8 vertici

float *g_field_t;
t_vector3d *grv_t;





// variabili degli spigoli:
t_vector3d  p[12]; // vertici interpolateti
t_vector3d gr[12]; // rispettivi gradienti interpolateti


int off;


// variabili di posizionamento-accesso:
float x, y, z; // posizione corrente nello spazio
float dx, dy, dz; // step fra i punti della spazio
float sx, sy, sz; // coordinate di partenza
float t; // buffer


g_num_iso_cells = 0;

sx = ox  -  dim_u *.5;
sy = oy  -  dim_v *.5;
sz = oz  -  dim_w *.5;

dx = dim_u / (float)nu;
dy = dim_v / (float)nv;
dz = dim_w / (float)nw;


glBegin(GL_TRIANGLES);


//
//   Numerazione dei vertici 
//   nell'algoritmo:
//  
//     4 ____________5
//     /|           /|
//  >7<____>______ 6 |
//   |            |  |
//   |  |         |  |
//   v 0|________ | _|1  FIELD 2
//   |  /         |  / 
//   |/___________|/   FIELD 1
//  3              2
//
// *Sistema di riferimento:*
//   .origine in v7
//   .orientamento:
//    _
//    z|
//   /
//  O - x - >
//  |
//  y
// \/
//  
//     -x> : 7-6>
//     -y> : 7-3>
//     -z> : 7-4>
//
//   .criterio di accesso:
//     for(z)
//      for(y)
//       for(x){}

// STEP 0/4:

// * * * Initial Setup: * * *
// Fills the first four XY plane with values:


g_field_0 = g_grid_vals + DIM_X*DIM_Y*1;
g_field_1 = g_grid_vals + DIM_X*DIM_Y*2; 
g_field_2 = g_grid_vals + DIM_X*DIM_Y*3;
g_field_3 = g_grid_vals + DIM_X*DIM_Y*4;

off = 0;
  for(iy=0, y=sy; iy < nv; iy++, y+=dy)
  {
   off = iy*GRID_DIM; // set the right offset for the current line

   for(ix=0, x=sx; ix<nu; ix++, x+=dx, off++)
   {
		 /*
     g_field_0[off] =  g_grid_vals[ix+iy*GRID_DIM]; // plane #0
      g_field_1[off] = g_grid_vals[ix+iy*GRID_DIM];         // plane #1
      g_field_2[off] = g_grid_vals[ix+iy*GRID_DIM];    // plane #2
     g_field_3[off] =  FIELD_FUN(x, y, sz + dz*2.0); // plane #3
		 */

//     4 ____________5
//     /|           /|
//   7<off>________ 6 |
//   |            |  |
//   |  |         |  |
//   v 0|________ | _|1  FIELD 2
//   |  /         |  / 
//   |/___________|/   FIELD 1
//  3              2
//
	 grv_1[off].x = g_field_1[off - 1]  -  g_field_1[off + 1];
	 grv_1[off].y = g_field_1[off - GRID_DIM]  -  g_field_1[off + GRID_DIM];
	 grv_1[off].z  = g_field_0[off]  -  g_field_2[off];

	 grv_2[off].x = g_field_2[off - 1]  -  g_field_2[off + 1];
	 grv_2[off].y = g_field_2[off - GRID_DIM]  -  g_field_2[off + GRID_DIM];
	 grv_2[off].z  = g_field_1[off]  -  g_field_3[off];

	//  grv_3[off].x = grv_0[off].y = grv_0[off].z = 
	//  grv_3[off].x = grv_3[off].y = grv_3[off].z = 0.0;

   }
  }


// STEP 1/4:

z = sz; // start from the 1nd plane:

for(iz=0; iz < nw; iz++, z+=dz)  // cycle through all the planes (back to front)
{

if(iz) // we are now at plane #1:
{
// go deeper 1 slice along the Z axis:
// [refresh XY plane pointers]

// rotate vertex value pointers:
 g_field_t = g_field_0;
  g_field_0 = g_field_1;
  g_field_1 = g_field_2;
  g_field_2 = g_field_3;
  g_field_3 = g_grid_vals + DIM_X*DIM_Y*(iz+4);

// rotate vertex gradient pointers:
 grv_t = grv_0;
  grv_0 = grv_1;
  grv_1 = grv_2;
  grv_2 = grv_3;
  grv_3 = grv_t;


  // evalutate the new deeper XY plane values
  for(iy=0, t=z+dz*2.0 , y=sy; iy<nv; iy++, y+=dy)
  {
   off = iy*GRID_DIM;
    for(ix=0, x=sx;  ix < nu;  ix++, x+=dx, off++)
    {
	  //g_field_3[off] = FIELD_FUN(x, y, t);

	   grv_1[off].x = g_field_1[off - 1]  -  g_field_1[off + 1];
	   grv_1[off].y = g_field_1[off - GRID_DIM]  -  g_field_1[off + GRID_DIM];
	   grv_1[off].z  = g_field_0[off]  -  g_field_2[off];

	   grv_2[off].x = g_field_2[off - 1]  -  g_field_2[off + 1];
	   grv_2[off].y = g_field_2[off - GRID_DIM]  -  g_field_2[off + GRID_DIM];
	   grv_2[off].z  = g_field_1[off]  -  g_field_3[off];

    }    
  }




}


/*
   Now we have to fetch the needed values into v0-7:
   [we are going to reuse the already computed cell values]
*/
  y = sy;

  off = 0;

for(iy=0; iy < nv; iy++, y+=dy )
{

   off = iy * GRID_DIM;

   x = sx;

 for(ix=0; ix < nu; ix++, x+=dx, off++)  
 {
 
  if( ix ) // we have already computed the field value of the first column	
  {
// Work fields are #1 & #2
//     4 _____<-_____5
//     /|           /|
//   7/____<-_____ 6 |
//   |            |  |
//   |  |         |  |
//   v 0|____<-__ | _|1   FIELD 2
//   |  /         |  / 
//   |/___<-______|/
//  3              2   FIELD 1
//

   // propagate previously memory-fetched values
   v[0] = v[1];
   v[3] = v[2];
   v[7] = v[6];
   v[4] = v[5];


   // propagate previously used vertex-gradient pointers:
   grv[0] = grv[1];
   grv[3] = grv[2];
   grv[7] = grv[6];
   grv[4] = grv[5];


  }
  else 
  {// fetch starting values from the array 
   
   // field values:
    v[0] = g_field_2[off + GRID_DIM];
    v[3] = g_field_1[off + GRID_DIM];
    v[4] = g_field_2[off];
    v[7] = g_field_1[off];

   // vertex-gradients:
    grv[0] = &grv_2[off + GRID_DIM];
    grv[3] = &grv_1[off + GRID_DIM];
    grv[7] = &grv_2[off];
    grv[4] = &grv_1[off];

  }


 v[1] = g_field_2[off+1 + GRID_DIM];
 v[2] = g_field_1[off+1 + GRID_DIM];
 v[5] = g_field_2[off+1];
 v[6] = g_field_1[off+1];

 grv[1] = &grv_2[off+1 + GRID_DIM];
 grv[2] = &grv_1[off+1 + GRID_DIM];
 grv[5] = &grv_2[off+1];
 grv[6] = &grv_1[off+1];




// STEP 3/4:

// Now all the 8 values have been fetched:


 /*
      Determine the index into the edge table which
      tells us which vertices are inside of the surface
 */
   cubeindex = 0x0000;
   if ( v[0] < iso ) cubeindex |= 0x0001;
   if ( v[1] < iso ) cubeindex |= 0x0002;
   if ( v[2] < iso ) cubeindex |= 0x0004;
   if ( v[3] < iso ) cubeindex |= 0x0008;
   if ( v[4] < iso ) cubeindex |= 0x0010;
   if ( v[5] < iso ) cubeindex |= 0x0020;
   if ( v[6] < iso ) cubeindex |= 0x0040;
   if ( v[7] < iso ) cubeindex |= 0x0080;


// Cube is entirely in/out of the surface:
if(cubeindex == 0   ) continue; // totally inside
if(cubeindex == 0xFF) continue; // totally outside
// ..else..
{


 edge_table_val = g_edge_table[ cubeindex ]; 



g_num_iso_cells++;
{

float x_dx = x + dx;
float y_dy = y + dy;
float z_dz = z + dz;

/* Find the vertices where the surface intersects the cube */


// GET_VAL_ISO( v_i, v_f, v_iso) 

//
//  VERTEX NUMBERING:
//     4 ____________5
//     /|           /|
//  >7<___________ 6 |
//   |            |  |
//   |  |         |  |
//   | 0|________ | _|1
//   |  /         |  / 
//   |/___________|/
//  3              2
//---------------------------
//  EDGE NUMBERING:
//       _____4______
//    7/|           5|
//  off____6______ / |
//   |            |  |
//  11  8        10 [9]
//   |  |___[0]__ | _|
//   | 3/         |  /[1]
//   |/___________|/
//          2      
//   note: edges 0, 1 and 9 will be computed for each cell
//   


//*** Compute interpolateted vertices and gradients on the 12 edges: ***
// [ use the bit-code in "edge_table_val" to check what edges contain a vertex ]

//////////////////
//* 0, 2, 4, 6 *//
//////////////////
   /*PIANO SOTTO: (y+dy) */


   /*PIANO SOPRA: (y) */
 
if (edge_table_val & 0x0040) //  edge 6
{
if(iz) // the 0-4-8-9 XY plane was already computed
{
 p[6]  = g_mem_p4_p[ off ];
 gr[6] = g_mem_gr4_p[ off ];
}
else
{

   if( iy ) // the upper line was already computed, so fetch the values for edge 6
   {
	p[6]  = g_mem_p2[ix];
	gr[6] = g_mem_gr2[ix];
   }
   else // we haven't already passed the first line:
   {
    t = GET_VAL_ISO(v[7], v[6], iso);

	p[6].x = x + dx*t;
	p[6].y = y;
	p[6].z = z;
    
    //eval_gradient(&p[6], &gr[6]);
	interpolate_vector_n(grv[7], grv[6], t,  &gr[6]);
   }

}


   }



  if (edge_table_val & 0x0010) //  edge 4
   {
   if(iy)
   {
    p[4]  = g_mem_p0_p[off - GRID_DIM];
	gr[4] = g_mem_gr0_p[off - GRID_DIM];
   }
    else
   {
    t = GET_VAL_ISO(v[4], v[5], iso);//?

	p[4].x = x + dx*t;
	p[4].y = y;
	p[4].z = z_dz;
    
//    eval_gradient(&p[4], &gr[4]);
	interpolate_vector_n(grv[4], grv[5], t,  &gr[4]);
   }

    g_mem_p4_p[off]  = p[4];
    g_mem_gr4_p[off] = gr[4];

   }




   if (edge_table_val & 0x0004) //  edge 2
   {

if(iz)
{
 p[2]  = g_mem_p0_p[off];
 gr[2] = g_mem_gr0_p[off];
}

else 
{
    t = GET_VAL_ISO(v[3], v[2], iso);//?

	p[2].x = x + dx*t;
	p[2].y = y_dy;
	p[2].z = z;
    
    // eval_gradient(&p[2], &gr[2]);
	interpolate_vector_n(grv[2], grv[3], t,  &gr[2]);
}


	g_mem_p2[ix]  = p[2];
	g_mem_gr2[ix] = gr[2];


   }

//*************************************
// edge 0 Always needs to be computed:
//*************************************
   if (edge_table_val & 0x0001)  //  edge 0
   {
    t = GET_VAL_ISO(v[0], v[1], iso);

	p[0].x = x + dx * t;
	p[0].y = y_dy;
	p[0].z = z_dz;
    
    //eval_gradient(&p[0], &gr[0]);
	interpolate_vector_n(grv[0], grv[1], t,  &gr[0]);

g_mem_p0_p[off]  = p[0];
g_mem_gr0_p[off] = gr[0];

    
   }
		                                     

//*******************************************************

///////////////////
//* 3, 7, 8, 11 *//
///////////////////

   if (edge_table_val & 0x0080) //  edge 7
   {

if(iy)
{
 gr[7] = g_mem_gr3[ix];
 p[7]  = g_mem_p3[ix];
}
else
{

    if(ix) 
    {
 	 gr[7] = gr[5];
	 p[7]  = p[5];
    }
    else
    {
     t = GET_VAL_ISO(v[7], v[4], iso);//?

	 p[7].x = x;
	 p[7].y = y;
	 p[7].z = z + dz * t;
    
     //eval_gradient(&p[7], &gr[7]);
	 interpolate_vector_n(grv[7], grv[4], t,  &gr[7]);
    }
}

   
   }



   if (edge_table_val & 0x0008) //  edge 3
   {
	if(ix)
	{
	 gr[3] = gr[1]; 
	 p[3]  = p[1];
	}
	 else
	{
     t = GET_VAL_ISO(v[3], v[0], iso);  //?

	 p[3].x = x;
	 p[3].y = y_dy;
	 p[3].z = z + dz*t;
    
    
	 //eval_gradient(&p[3], &gr[3]);
	 interpolate_vector_n(grv[3], grv[0], t,  &gr[3]);
	 
	}


	g_mem_p3[ix] = p[3];
    g_mem_gr3[ix] = gr[3];

   }


   /*VERTICALI:*/




   if (edge_table_val & 0x0800) //  edge 11
   {

if(iz)
{
 p[11]  = g_mem_p8_p[off];
 gr[11] = g_mem_gr8_p[off];
}
else
{

   if(ix)
   {
    gr[11] = gr[10];
    p[11] = p[10];
   }
    else
   {
    t = GET_VAL_ISO(v[7], v[3], iso);//?

	p[11].x = x;
	p[11].y = y + dy*t;
	p[11].z = z;
    
    //eval_gradient(&p[11], &gr[11]);
	interpolate_vector_n(grv[7], grv[3], t,  &gr[11]);
   }
}
}


   if (edge_table_val & 0x0100) //  edge 8
   {

   if(ix)
   {
	gr[8] = gr[9];
	p[8]  = p[9];
   }
    else   
   {
    t = GET_VAL_ISO(v[4], v[0], iso);//?

	p[8].x = x;
	p[8].y = y + dy*t;
	p[8].z = z_dz;
    
    //eval_gradient(&p[8], &gr[8]);
	interpolate_vector_n(grv[4], grv[0], t,  &gr[8]);
   }

    g_mem_p8_p[off]  = p[8];
    g_mem_gr8_p[off] = gr[8];

   }



///////////////////
//* 1, 5, 9, 10 *//
///////////////////

   if (edge_table_val & 0x0020) //  edge 5
   {
	if(iy)
	{
	 gr[5] = g_mem_gr1[ix];
	 p[5]  = g_mem_p1[ix];
	}
	 else 
	{

    t = GET_VAL_ISO(v[6], v[5], iso);//?

	p[5].x = x_dx;
	p[5].y = y;
	p[5].z = z + dz*t;
    
    //eval_gradient(&p[5], &gr[5]);
	interpolate_vector_n(grv[6], grv[5], t,  &gr[5]);
	}

   }

//*************************************
// edge 1 Always needs to be computed:
//*************************************
   if (edge_table_val & 0x0002)  //  edge 1
   {
    t = GET_VAL_ISO(v[2], v[1], iso);//?

	p[1].x = x_dx;
	p[1].y = y_dy;
	p[1].z = z + dz * t;
    
    //eval_gradient(&p[1], &gr[1]);
	interpolate_vector_n(grv[1], grv[2], t,  &gr[1]);

	g_mem_p1[ix] = p[1];
    g_mem_gr1[ix] = gr[1];
   }



   if (edge_table_val & 0x0400) //  edge 10
   {

if(iz)
{
 p[10]  = g_mem_p9_p[off];
 gr[10] = g_mem_gr9_p[off];
}
 else 
{
    t = GET_VAL_ISO(v[6], v[2], iso);//?

	p[10].x = x_dx;
	p[10].y = y + dy*t;
	p[10].z = z;
    
    //eval_gradient(&p[10], &gr[10]);
	interpolate_vector_n(grv[6], grv[2], t,  &gr[10]);
}


}



//*************************************
// edge 9 Always needs to be computed:
//*************************************
   if (edge_table_val & 0x0200) //  edge 9
   {
    t = GET_VAL_ISO(v[5], v[1], iso);//?

	p[9].x = x_dx;
	p[9].y = y + dy*t;
	p[9].z = z_dz;
    
    //eval_gradient(&p[9], &gr[9]);
	interpolate_vector_n(grv[5], grv[1], t,  &gr[9]);

    g_mem_p9_p[off]  = p[9];
    g_mem_gr9_p[off] = gr[9];
   }






//*************************************//

// STEP 4/4:




   /* Ora che abbiamo i punti e i gradienti interpolateti 
      possiamo disegnare i triangoli pesenti nella cella */

   for(i = 0; g_iso_tri_table[cubeindex][i] != -1; i+=3 ) 
    {
/*
{   
	  t_vector3d n, v_a, v_b;
	  int i1, i2, i3;

	  i1 = g_iso_tri_table[cubeindex][i];
	  i2 = g_iso_tri_table[cubeindex][i+2];
      i3 = g_iso_tri_table[cubeindex][i+1];
	  vector3d_make((t_point3d *)&p[i1], (t_point3d *)&p[i2], &v_a);
      vector3d_make((t_point3d *)&p[i1], (t_point3d *)&p[i3], &v_b);
      vector3d_prod_v(&v_a, &v_b,  &n);
	  vector3d_norm( &n );

	  glNormal3fv( &n.x );
      glVertex3fv( &p[i1].x );
      glVertex3fv( &p[i2].x );
      glVertex3fv( &p[i3].x );
}
*/
      ind = g_iso_tri_table[ cubeindex ][i];
// vector3d_norm(&gradiente_ris[ind]);
		glTexCoord1f(gr[ind].x);
	  glNormal3fv( &gr[ind].x );
      glVertex3fv( &p[ind].x );


     ind = g_iso_tri_table[ cubeindex ][i+2];
// vector3d_norm(&gradiente_ris[ind]);
		 glTexCoord1f(gr[ind].x);
	  glNormal3fv( &gr[ind].x );
      glVertex3fv( &p[ind].x );


	 ind = g_iso_tri_table[ cubeindex ][i+1];
// vector3d_norm(&gradiente_ris[ind]);
	 glTexCoord1f(gr[ind].x);
	  glNormal3fv( &gr[ind].x );
      glVertex3fv( &p[ind].x );

	  num_tri++;
    }






} // end __inline Polygonise();


  } // for(ix)

} // if(cubeindex!=0xFF){}



 } // for(iy)
   

} // for(iz)


glEnd();


return num_tri;
}





#endif

